/*
    Настройки Arduino IDE
    1. Установить Arduino-Pico, как описано тут:
       https://arduino-pico.readthedocs.io/en/latest/install.html
    2. Выбрать процессор (Waveshare RP2040 Zero).
    3. Немного его "подразогнать" (установить CPU Clock из числа
       тех, у которых приписка (Overclock). Я ставил до 300 MHz.
*/

/* Число Pi */
const double valueOfPi = 3.1415926535;

/*
    ----------------- настройки для клавиатуры --------------------
    pin_0 - 0-й разряд, к нему подключена кнопка запуска калибровки
    pin_1 - 1-й разряд, к нему подключена кнопка переключения измерения L или C
    Когда нажата кнопка "переключения измерения L или C" сканкод клавиатуры будет 0b00000001
    Когда нажата кнопка "запуска калибровки" сканкод клавиатуры будет 0b00000010
*/
#define KEYBOARD_PIN_0          3
#define KEYBOARD_PIN_1          4
#define SUCCESS_HITS_NUMBER     6
#define KEYBOARD_MASK           0b00000011
int keyboardHitCounter =        0;
uint8_t keyboardPreviousCode =  KEYBOARD_MASK;
bool buttonPressedFlag =        false;
bool waitButtonRelease =        false;

#include <hardware/pwm.h>
uint signalSlice;
volatile uint32_t signalWrapCount = 0;
/* Пин входа частотомера */
#define SIGNAL_PIN 1

/*
   Настройки и переменные для двух таймеров. Первый таймер срабатывает каждые 10 мс.
   и используется для обработки клавиатуры. Второй таймер срабатывает каждую секунду
   и используется для подсчета частоты сигнала, поступающего от измерительного
   генератора.
*/
#include <hardware/timer.h>
#include <hardware/irq.h>
/* первый таймер */
#define ALARM_NUM0 1
#define ALARM_IRQ0 TIMER_IRQ_1
uint32_t currentAlarm;
#define  TIMER0_INTERRUPT_PERIOD 10000
bool timer0IntFlag = false;
/* второй таймер */
#define ALARM_NUM1 2
#define ALARM_IRQ1 TIMER_IRQ_2
uint32_t currentAlarm1;
#define  TIMER1_INTERRUPT_PERIOD 1000000
bool oneSecondFlag = false;
int oneSecondCounter = 0;

/* емкость "рефренсного" конденсатора (тут нужна максимальная точность!) */
double ref_capacitance = 21.353E-9;
/* емкость рефренсной катушки индуктивности (тут точность не особо критична, т.к.
  значение переписывается программой после выполнения процедуры калибровки) */
double ref_inductance = 44.299E-6;

/* остальные переменные */
uint8_t currentMode = 0;
uint8_t rememberCurrentMode = 0;
uint8_t currentDisplay = 8 ;
double valueToDisplay;
String component, unit1, unit1000 , unit1000000;
bool rememberRelayState = false;
/* сетчик и измеренная частота */
volatile uint32_t currentFrequency, calibrationFrequency;
volatile uint16_t counter_lsb , counter_msb;

/* пин, который управляет реле переключения измерения L или C */
#define relayControlPin       0

/*
    настройки для управления многоцветным светодиодом (он меняет цвет свечения
    в зависимости от текущего режима измерения)..
*/
#include <Adafruit_NeoPixel.h>
#define PIN        16   // Пин, к которому подключен многоцветный светодиод
#define NUMPIXELS 1     // Сколько светодиодов в цепи всего
Adafruit_NeoPixel pixels(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);

/* настройки для работы с графическим ЖКИ-дисплеем */
#include <Adafruit_GFX.h>       // установка с сайта github -> https://github.com/adafruit/Adafruit-GFX-Library
#include <Adafruit_ST7735.h>    // установка с сайта github -> https://github.com/adafruit/Adafruit-ST7735-Library
#include <SPI.h>
/* параметры и пины, используемые для подключения графического ЖКИ-дисплея */
#define TFT_RST    28    // нога Reset
#define TFT_DC     27    // нога Data / Command
#define TFT_CS     -1    // нога Chip Select (подключено к GPIO 5, но при устнаовке "-1" ей автоматически управляет само "железо")
Adafruit_ST7735 * tft;

/* setup для первого ядра */
void setup() {
  pixels.begin(); // INITIALIZE NeoPixel strip object (REQUIRED)
  setLedRed();
  /*------------ Настройка прерывания таймера ---------------------------*/
  hw_set_bits(&timer_hw->inte, 1u << ALARM_NUM0);
  irq_set_exclusive_handler(ALARM_IRQ0, alarm_irq0);
  irq_set_enabled(ALARM_IRQ0, true);
  currentAlarm = (uint32_t)(timer_hw->timerawl + TIMER0_INTERRUPT_PERIOD);
  timer_hw->alarm[ALARM_NUM0] = currentAlarm;
  /*---------------------------------------------------------------------*/
  /* инициализация дисплея */
  /* сначала переназначаем номера пинов для сигналов MISO и SCK шины SPI(0) */
  SPI.setTX(7);
  SPI.setSCK(6);
  SPI.begin(true); // "true" обозначает аппаратное управление пином CS
  /* и лишь потом непосредственно инициализируем дисплей */
  tft = new Adafruit_ST7735(&SPI, TFT_CS, TFT_DC, TFT_RST);
  tft->initR( INITR_BLACKTAB );
  tft->setRotation( 1 );
  /* настройка выхода, используемого для управления реле */
  digitalWrite(relayControlPin, LOW);
  pinMode(relayControlPin, OUTPUT);
  /* настройка входов, к которым подключены кнопки управления */
  pinMode(KEYBOARD_PIN_0, INPUT_PULLUP);
  pinMode(KEYBOARD_PIN_1, INPUT_PULLUP);
  /* зажигаем светодиод зеленым цветом */
  setLedGreen();
}

/* setup для второго ядра */
void setup1() {
  /*------------ Настройка прерывания таймера ---------------------------*/
  hw_set_bits(&timer_hw->inte, 1u << ALARM_NUM1);
  irq_set_exclusive_handler(ALARM_IRQ1, alarm_irq1);
  irq_set_enabled(ALARM_IRQ1, true);
  currentAlarm1 = (uint32_t)(timer_hw->timerawl + TIMER1_INTERRUPT_PERIOD);
  timer_hw->alarm[ALARM_NUM1] = currentAlarm1;
  /*---------------------------------------------------------------------*/
  /* настраиваем счетчик PWM на подсчет числа входных импульсов */
  signalSlice = setup_pwm_counter(SIGNAL_PIN);
  /* активируем прерывания */
  interrupts();
}

/* основной цикл для второго ядра */
void loop1() {
}

/* основной цикл для первого ядра */
void loop() {
  /*
     Каждые 10 миллисекунд (по флагу таймера 0) сканируем клавиатуру.
     Если кнопка была нажата, вызываем процедуру "анализза", после
     чего сбрасываем флаг(и)
  */
  if (timer0IntFlag) {
    readKeyboardData();
    if (buttonPressedFlag) {
      analizeKeyboard();
      buttonPressedFlag = false;
    }
    timer0IntFlag = false;
  }

  /* действия, совершаемые раз в секунду */
  if (oneSecondFlag) {
    currentFrequency = ( 65536 * counter_msb ) + counter_lsb;
    if (currentMode > 2) {
      exaclyCalibration();
    }
    if (Serial) {
      Serial.println(currentFrequency);
    }
    displayInformation();
    oneSecondFlag = false;
  }
}

/* обработчик прерывания 10-ти миллисекундного таймера */
static void alarm_irq0(void) {
  /* ставим флаг срабатывания таймера 0 */
  timer0IntFlag = true;
  /* сбрасываем флаг прерывания */
  hw_clear_bits(&timer_hw->intr, 1u << ALARM_NUM0);
  /* увеличиваем время, когда произойдет следующее прерывание на период, с которым должны происходить прерывания */
  currentAlarm += TIMER0_INTERRUPT_PERIOD;
  /* записываем в таймер время, когда должно будет произойти следующее прерыввание */
  timer_hw->alarm[ALARM_NUM0] = currentAlarm;
}

/* обработчик прерывания 1-но секундного таймера */
static void alarm_irq1(void) {
  /*
    из числа переполнений счетчика PWM и текущего значения счетчика PWM вычисляем
    значение частоты сигнала, поступающего на вход счетчика PWM 
  */
  /* сохраняем значение счетчика PWM в промежуточный регистр */
  counter_lsb = pwm_get_counter(signalSlice);
  /* сбрасываем на ноль счетчик PWM */
  pwm_set_counter(signalSlice, 0);
  /* сохраняем значение счетчика числа переполнений PWM в промежуточный регистр */
  counter_msb = signalWrapCount;
  /* сбрасываем на ноль счетчик числа переполнений PWM */
  signalWrapCount = 0;
  /* ставим флаг "одна секунда" (для других задач осовного цикла) */
  oneSecondFlag = true;
  /* сбрасываем флаг прерывания */
  hw_clear_bits(&timer_hw->intr, 1u << ALARM_NUM1);
  /* увеличиваем время, когда произойдет следующее прерывание на период, с которым должны происходить прерывания */
  currentAlarm1 += TIMER1_INTERRUPT_PERIOD;
  /* записываем в таймер время, когда должно будет произойти следующее прерыввание */
  timer_hw->alarm[ALARM_NUM1] = currentAlarm1;
}

/*
   "анализ клавиатуры" - вызов нужной процедуры в зависимости от полученного сканкода
*/
void analizeKeyboard() {
  switch (keyboardPreviousCode) {
    case 0b00000010:
      startOfCalibration();
      break;
    case 0b00000001:
      switchMode();
      displayInformation();
      break;
    case 0b00000000:
      Serial.begin(115200);
      Serial.println("Serial port enabled");
      break;
  }
}

/* вывод на ЖКИ-дисплей показаний в зависимости от "текущего режима" (currentMode)*/
void displayInformation() {
  switch (currentMode) {

    /* отображение показаний индуктивности */
    case 0:
      valueToDisplay = getValueOf(ref_capacitance, ref_inductance, currentFrequency, 1E6);
      unit1 = " uH";
      unit1000 = " mH";
      unit1000000 = " H";
      component = "inductor";
      if (currentFrequency < 100) {
        showNotConnected();
        break;
      }
      if (currentDisplay != currentMode ) {
        tft->fillScreen(ST77XX_BLACK);
        tft->drawRect(2, 2, 156, 124, 0xAD35);
        drawTexxt( 50, 85, 100, ST77XX_BLACK, ST77XX_RED, 1, "L1=" + String( ref_inductance * 1E6) + " uH" );
        currentDisplay = currentMode;
      }
      showExaclyValue();
      break;

    /* отображение показаний емкости */
    case 1:
      valueToDisplay = getValueOf(ref_inductance, ref_capacitance, currentFrequency, 1E12);
      unit1 = " pF";
      unit1000 = " nF";
      unit1000000 = " uF";
      component = "capasitor";
      if (currentFrequency < 100 or currentFrequency > 200000 or valueToDisplay < -1.0 ) {
        showNotConnected();
        break;
      }
      if (currentDisplay != currentMode ) {
        tft->fillScreen(ST77XX_BLACK);
        tft->drawRect(2, 2, 156, 124, 0xAD35);
        drawTexxt( 50, 85, 100, ST77XX_BLACK, ST77XX_RED, 1, "L1=" + String( ref_inductance * 1E6) + " uH" );
        currentDisplay = currentMode;
      }
      showExaclyValue();
      break;
  }
}

/* Вывод на дисплей слова Wait (ждите) */
void showWait() {
  if (currentDisplay != 3) {
    tft->fillScreen(ST77XX_RED);
    tft->drawRect(2, 2, 156, 124, ST77XX_BLACK);
    drawTexxt( 50, 54, 100, ST77XX_RED, ST77XX_BLACK, 3, "WAIT" );
    currentDisplay = 3;
  }
}

/* Вывод на дисплей сообщения, что компонент не подключен */
void showNotConnected() {
  if (currentDisplay != 2) {
    tft->fillScreen(ST77XX_RED);
    tft->drawRect(2, 2, 156, 124, ST77XX_BLACK);
    drawTexxt( 35, 46, 120, ST77XX_RED, ST77XX_YELLOW, 1, "Connect" );
    drawTexxt( 35, 60, 120, ST77XX_RED, ST77XX_YELLOW, 1, component );
    drawTexxt( 35, 74, 120, ST77XX_RED, ST77XX_YELLOW, 1, "to measure" );
    currentDisplay = 2;
  }
}

/* вывод на дисплей непосредственно результатов измерений */
void showExaclyValue() {
  if ( valueToDisplay >= 1000000 ) {
    drawTexxt( 31, 35, 100, ST77XX_BLACK, ST77XX_GREEN, 2, String( valueToDisplay / 1000000 ) + unit1000000 );
  } else if ( valueToDisplay >= 1000 ) {
    drawTexxt( 31, 35, 100, ST77XX_BLACK, ST77XX_GREEN, 2, String( valueToDisplay / 1000 ) + unit1000 );
  } else {
    drawTexxt( 31, 36, 100, ST77XX_BLACK, ST77XX_GREEN, 2, String( valueToDisplay ) +  unit1 );
  }
  drawTexxt( 20, 64, 120, ST77XX_BLACK, ST77XX_YELLOW, 1, "Freq : " + String( currentFrequency ) + " Hz" );
}

/* процедура вывода на дисплей одной строки (любого) текста */
void drawTexxt( int16_t x, int16_t y, int16_t maxTextWidth, int bgColor, int txtColor, uint8_t fontSize, String text )
{
  int pos = (160 - ( text.length() * 6 * fontSize )) / 2;          // вычисляем стартовую позицию для текста в строке (с учетом размещения текста по центру)
  tft->fillRect( 3, y - 2 , 154, ( 8 * fontSize ) + 3, bgColor );  // "закрашиваем" строку цветом фона
  tft->setCursor( pos, y );                                        // устанавливаем курсор на позицию старта строки
  tft->setTextColor( txtColor );                                   // устанавливаем цвет текста
  tft->setTextWrap( false );                                       // выключаем перенос сроки
  tft->setTextSize(fontSize);                                      // устанавливаем размер шрифта
  tft->print( text );                                              // выводим текст
}

/* процедура вычисления одного из номиналов контура (любого) по измеренной частоте и рефренсному номиналу второго */
float getValueOf(double anotherValue, double refrenceValue, uint32_t mesuredFreq, double multiplaxer) {
  float toReturn = (1.0 / ( 4.0 * mesuredFreq * mesuredFreq * anotherValue * valueOfPi * valueOfPi )) - refrenceValue;
  toReturn *= multiplaxer;
  return toReturn;
}

/* процедура переключения режима измерения с L на C и обратно */
void switchMode() {
  switch (currentMode) {
    case 0:                                    // если был режим "0" (измерение индуктивности)...
      currentMode = 1;                         // ...то включаем режим "1" (измерение емкости)
      digitalWrite( relayControlPin, HIGH );   // ... и подаем сигнал для включения реле
      setLedBlue();
      break;
    case 1:                                    // если был режим "1" (измерение емкости)...
      currentMode = 0;                         // ...то включаем режим "0" (измерение индуктивности)
      digitalWrite( relayControlPin, LOW );    // ... и выключаем реле
      setLedGreen();
      break;
  }
}

/* запуск процедуры калибровки прибора */
void startOfCalibration() {
  setLedRed();
  rememberCurrentMode = currentMode;                  // запоминаем режим, из которого запускалась калибровка
  rememberRelayState = digitalRead(relayControlPin);  // запоминаем состояние выхода управления реле "R или C"
  currentMode = 3;                                    // переключаем текущий режим на старт "калибровки" (3) и
  exaclyCalibration();                                // ... вызываем реальную процедуру калибровки
}

/* непосредственно сама процедура калибровки */
void exaclyCalibration() {
  switch (currentMode) {
    case 3:
      digitalWrite( relayControlPin, HIGH );
      showWait();
      currentMode++;
      break;
    case 4:
      calibrationFrequency = 0;
      currentMode++;
      break;
    case 5:
      calibrationFrequency += currentFrequency;
      currentMode++;
      break;
    case 6:
      calibrationFrequency += currentFrequency;
      calibrationFrequency = calibrationFrequency / 2 ;
      currentMode++;
      break;
    case 7:
      calibrationFrequency += currentFrequency;
      calibrationFrequency = calibrationFrequency / 2 ;
      ref_inductance = getValueOf(ref_capacitance, 0.0, calibrationFrequency, 1.0);
      digitalWrite( relayControlPin, rememberRelayState );
      currentMode = rememberCurrentMode;
      if (currentMode == 1) {
        setLedBlue();
      } else {
        setLedGreen();
      }
      break;
  }
}

/* сканирование клавиатуры, защита от дребезга, задержка после отпускания */
void readKeyboardData() {
  /* считываем текущее состояние пинов клавиатуры */
  byte NewScanCode = (digitalRead(KEYBOARD_PIN_1) << 1) | digitalRead(KEYBOARD_PIN_0);
  /* отбрасываем лишние биты по маске (на всякий случай) */
  NewScanCode &= KEYBOARD_MASK;
  /* эта проверка определяет задержку между отпусканием предыдущей кнопки и нажатием следующей */
  if (keyboardPreviousCode == KEYBOARD_MASK && keyboardHitCounter < (10 * SUCCESS_HITS_NUMBER)) {
    keyboardHitCounter++ ;
    return;
  }
  /* если старый и новый сканкоды совпали, */
  if (NewScanCode == keyboardPreviousCode) {
    /* .. просто увеличиваем счетчик совпадений сканкодов */
    keyboardHitCounter++ ;
  } else {
    keyboardPreviousCode = NewScanCode;  // "новый" сканкод становится "старым"
    keyboardHitCounter = 0;              // счетчик совпадений сбрасывается
    buttonPressedFlag = false;           // флаг "была нажата кнопка" сбрасывается (чисто на всякий случай)
  }
  /* если число повторов равно заданному (кнопка была нажата дольше времени подавления дребезга) */
  if ( keyboardHitCounter == SUCCESS_HITS_NUMBER) {
    /* устанавливаем флаг "была нажата кнопка" */
    buttonPressedFlag = true;
  }
  /* этот "костыль" не даст переполнится счетчику повторов, сколько бы не держали кнопку нажатой */
  if (keyboardHitCounter > SUCCESS_HITS_NUMBER) {
    keyboardHitCounter --;
  }
}

/* процедуры установки цвета свечения многоцветного светодиода */
/* красный */
void setLedRed() {
  pixels.setPixelColor(0, pixels.Color(20, 0, 0));
  pixels.show();   // Send the updated pixel colors to the hardware.
}
/* зеленый */
void setLedGreen() {
  pixels.setPixelColor(0, pixels.Color(0, 20, 0));
  pixels.show();   // Send the updated pixel colors to the hardware.
}
/* синий */
void setLedBlue() {
  pixels.setPixelColor(0, pixels.Color(0, 0, 20));
  pixels.show();   // Send the updated pixel colors to the hardware.
}

/* настройка и инициализация счетчика PWM, а также его прерываний */
uint setup_pwm_counter(uint gpio) {
  // can only use input on channel B
  assert(pwm_gpio_to_channel(gpio) == PWM_CHAN_B);

  uint slice_num = pwm_gpio_to_slice_num(gpio);
  pwm_config cfg = pwm_get_default_config();
  pwm_config_set_clkdiv_mode(&cfg, PWM_DIV_B_RISING);
  pwm_config_set_clkdiv(&cfg, 1.f);
  pwm_init(slice_num, &cfg, false);
  gpio_set_function(gpio, GPIO_FUNC_PWM);

  // register interrupt handler
  pwm_clear_irq(slice_num);
  pwm_set_irq_enabled(slice_num, true);
  irq_set_exclusive_handler(PWM_IRQ_WRAP, on_pwm_wrap);
  irq_set_enabled(PWM_IRQ_WRAP, true);

  pwm_set_enabled(slice_num, true);
  return slice_num;
}

/* обработчик прерывания переполнения счетчика PWM */
void on_pwm_wrap() {
  pwm_clear_irq(signalSlice);
  signalWrapCount++;
}
